\section{Context}

> {-# LANGUAGE DeriveFunctor, DeriveFoldable, DeriveTraversable,
>              TypeOperators, FlexibleInstances #-}

> module Context where

> import Prelude hiding (any)
> import Control.Monad.Error ()
> import Control.Monad.State
> import Data.Foldable
> import Data.Traversable

> import Basics
> import Group


We use a fixed collection |BaseUnit| of base units to avoid adding an
extra parameter to everything.

> data BaseUnit = METRE | SEC | KG
>     deriving (Eq, Ord, Show)

> type Dim a = GExp BaseUnit a


Types are parameterised over a set of variables. We use a single set
for both sorts, to simplify operations like binding variables in type
schemes.

> data Ty a  =  V a
>            |  Ty a :-> Ty a
>            |  Float (Dim a)
>     deriving (Eq, Show, Functor, Foldable, Traversable)
> infixr 5 :->


Variables are assigned sorts and properties in |Prop|.

> data Prop a  =  TY (Maybe (Ty a))
>              |  GE (Maybe (Dim a))
>     deriving (Eq, Show, Functor, Foldable, Traversable)


A type scheme is a collection of bindings wrapping a type. Each
binding has a property giving its sort, and |Nothing| for $\forall$ or
|Just t| for a definition. Note that we use a nested representation to
bind a variable: see ``de Bruijn notation as a nested data type''
(Bird and Paterson 1999).

> data Index a = Z | S a
>     deriving (Eq, Ord, Show, Functor, Foldable, Traversable)

> data Schm a  =  Type (Ty a) 
>              |  Bind (Prop a) (Schm (Index a))
>     deriving (Eq, Show, Functor, Foldable, Traversable)


We use |:=| as a king of fancy pair that stores a variable and its
property; sometimes this will be of a known sort.

> data a := b = a := b
>     deriving (Eq, Show, Functor, Foldable, Traversable)


Terms are parameterised over a set of variables, but do not use a
nested representation, because we will do scope checking along with
type checking.

> data Tm v  =  X v
>            |  Tm v :$ Tm v 
>            |  Lam v (Tm v)
>            |  Let v (Tm v) (Tm v)
>     deriving (Eq, Show)


Let's have fancy pairs for term variables and their schemes, while
we're at it.

> data a :::  b  = a :::  b
>     deriving (Eq, Show)


Now we can define context entries, as (type or dimension) variable and
property, term variable and scheme, or $\fatsemi$ marker.

> data Entr a   =   A (a := Prop a)
>                |  T (TmName ::: Schm a)
>                |  LetGoal
>     deriving (Eq, Show)


Let there be type synonyms! Integers are handy for type variable names
as we need to generate fresh ones, but we can use strings for term
variables since we just need to compare them for equality.

> type TyName         = Integer
> type TmName         = String

> type Dimension      = Dim   TyName
> type Type           = Ty    TyName
> type Scheme         = Schm  TyName
> type Term           = Tm    TmName
> type Entry          = Entr  TyName
> type Property       = Prop  TyName
> type Declaration    = TyName := Property
> type DeclarationTY  = TyName := Maybe Type
> type DeclarationGE  = TyName := Maybe Dimension

> type TmEntry        = TmName ::: Scheme


The |FV| typeclass allows us to test whether a type variable name
(integer) occurs within a value. Integers themselves form the base
case, then thanks to |-XFlexibleInstances| we can test membership in
any |Foldable| structure.

> class FV a where
>     (<?) :: TyName -> a -> Bool

> instance FV Integer where
>     (<?) = (==)

> instance (Foldable t, FV a) => FV (t a) where
>     alpha <? t = any (alpha <?) t


Contexts contain all sorts of entries, while suffixes only contain
type and group variables declarations.

> type Context  = Bwd Entry
> type Suffix   = Fwd Declaration


The `fish' operator lets us stick a suffix on the end of a context.

> (<><) :: Context -> Suffix -> Context
> _Gamma <>< F0               = _Gamma
> _Gamma <>< (alphaD :> _Xi)  = _Gamma :< A alphaD <>< _Xi
> infixl 8 <><



\subsection{The |Contextual| monad}

When implementing the algorithms, we work in a monad that carries some
state (the next fresh name and the current context) and allows failure
with an error message.

> type Contextual  = StateT (TyName, Context) (Either String)


Some handy monadic commands for fresh variable generation: |freshName|
returns the current fresh name and increments the counter, and |fresh|
adds a fresh variable to the context with the given property.

> freshName :: Contextual TyName
> freshName = do  (beta, _Gamma) <- get
>                 put (succ beta, _Gamma)
>                 return beta

> fresh :: Property -> Contextual TyName
> fresh d = do  beta <- freshName
>               modifyContext (:< A (beta := d))
>               return beta


These commands specialise the state monad machinery to work on the
context component.

> getContext :: Contextual Context
> getContext = gets snd
>
> putContext :: Context -> Contextual ()
> putContext _Gamma = do  beta <- gets fst
>                         put (beta, _Gamma)
>
> modifyContext :: (Context -> Context) -> Contextual ()
> modifyContext f = getContext >>= putContext . f


A common pattern in the algorithms is searching the context for
variable declarations (perhaps of a particular sort), making some
changes and either choosing to |restore| the existing declaration or
|replace| it with a new one. The |onTop| function captures this
pattern, along with its single-sorted variants |onTopTY| and
|onTopGE|.

> data Extension = Restore | Replace Suffix

> restore :: Contextual Extension
> restore = return Restore

> replace :: Suffix -> Contextual Extension
> replace = return . Replace

> replaceOne :: Declaration -> Contextual Extension
> replaceOne = replace . (:> F0)

> onTop :: (Declaration -> Contextual Extension) -> Contextual ()
> onTop f = do
>     _Gamma :< vD <- getContext
>     putContext _Gamma
>     case vD of
>         A alphaD    -> do  m <- f alphaD
>                            case m of
>                                Replace _Xi  -> modifyContext (<>< _Xi)
>                                Restore      -> modifyContext (:< vD)
>         _            -> onTop f >> modifyContext (:< vD)

> onTopTY :: (DeclarationTY -> Contextual Extension) -> Contextual ()
> onTopTY f = onTop $
>     \ (gamma := d) -> case d of
>         TY t  -> f (gamma := t)
>         _     -> onTopTY f >> restore

> onTopGE :: (DeclarationGE -> Contextual Extension) -> Contextual ()
> onTopGE f = onTop $
>     \ (gamma := d) -> case d of
>         GE t  -> f (gamma := t)
>         _     -> onTopGE f >> restore