\section{Unify}

> module Unify (unify) where

> import Prelude hiding (mapM_)
> import Data.Foldable (mapM_)

> import Basics
> import Context
> import GroupUnify

> unify :: Type -> Type -> Contextual ()
> unify (tau0 :-> tau1)  (upsilon0 :-> upsilon1)  = unify tau0 upsilon0 >> unify tau1 upsilon1
> unify (Float d)       (Float e)               = unifyGExp d e
> unify (V alpha)        (V beta)                 = onTopTY $
>   \ (gamma := d) -> case
>           (gamma == alpha,  gamma == beta,  d         ) of
>           (True,            True,           _         )  ->  restore                                 
>           (True,            False,          Nothing   )  ->  replaceOne $ alpha := TY (Just (V beta))
>           (False,           True,           Nothing   )  ->  replaceOne $ beta := TY (Just (V alpha))
>           (True,            False,          Just tau  )  ->  unify (V beta)   tau       >> restore   
>           (False,           True,           Just tau  )  ->  unify (V alpha)  tau       >> restore   
>           (False,           False,          _         )  ->  unify (V alpha)  (V beta)  >> restore   
> unify (V alpha)        tau                               =   startSolve alpha tau
> unify tau              (V alpha)                         =   startSolve alpha tau    
> unify _                _                                 =   fail "Rigid-rigid mismatch!"

Something interesting happens here: when we start to solve a
flex-rigid constraint, we have to be careful not to throw away
polymorphism. The syntactic occurs check performed by |solve| is not
enough, because the richer equational theory of abelian groups may
exhibit apparent dependency when there is in fact none. To deal with
this, we replace group expressions in the rigid type with fresh
variables, solve the flex-rigid constraint first, then unify the group
expressions. Note that we have to do a similar thing whenever adding a
type variable definition to the suffix.

> startSolve :: TyName -> Type -> Contextual ()
> startSolve alpha tau = do
>     (rho, xs) <- rigidHull tau
>     solve alpha (constraintsToSuffix xs) rho
>     solveConstraints xs
>     return ()

> rigidHull :: Type -> Contextual (Type, Fwd (TyName, Dimension))
> rigidHull (V a)              = return (V a, F0)
> rigidHull (tau :-> upsilon)  = do  (tau',      xs  )  <- rigidHull tau
>                                    (upsilon',  ys  )  <- rigidHull upsilon
>                                    return (tau' :-> upsilon', xs <+> ys)
> rigidHull (Float d)          = do  beta <- freshName
>                                    return (Float (embedVar beta), (beta, d) :> F0)

> constraintsToSuffix :: Fwd (TyName, Dimension) -> Suffix
> constraintsToSuffix = fmap ((:= GE Nothing) . fst)

> solveConstraints :: Fwd (TyName, Dimension) -> Contextual ()
> solveConstraints = mapM_ (uncurry $ unifyGExp . embedVar)


> solve :: TyName -> Suffix -> Type -> Contextual ()
> solve alpha _Xi tau = onTopTY $
>   \ (gamma := d) ->
>     let occurs = gamma <? tau || gamma <? _Xi
>     in case (gamma == alpha,  occurs,  d             ) of
>             (True,            True,    _             )  ->  fail "Occurrence detected!"
>             (True,            False,   Nothing       )  ->  replace $ _Xi <+> ((alpha := TY (Just tau)) :> F0)
>             (True,            False,   Just upsilon  )  ->  modifyContext (<>< _Xi)
>                                                         >>  unify upsilon tau
>                                                         >>  restore
>             (False,           True,    Nothing       )  ->  solve alpha ((gamma := TY d) :> _Xi) tau
>                                                         >>  replace F0
>             (False,           True,    Just upsilon  )  ->  do
>                 (upsilon', xs) <- rigidHull upsilon
>                 solve alpha (constraintsToSuffix xs <+> ((gamma := TY (Just upsilon')) :> _Xi)) tau
>                 solveConstraints xs
>                 replace F0   
>             (False,           False,   _             )  ->  solve alpha _Xi tau
>                                                         >>  restore